#!/usr/bin/perl
#
# log2graph.pl: generates a .dot file from fraglet execution traces
#
# Copyright (C) 2007 Lidia Yamamoto and Christian Tschudin
# <lidia.yamamoto@unibas.ch>, <christian.tschudin@unibas.ch>
#
# This file is distributed under the terms of the GNU general Public
# License (GPL), see the file COPYING
#
# Parses a fraglet execution trace in stdin, and outputs a file in
# .dot format containing all the species and reactions that occurred
# during the execution; this file can be read by the "neato" tool from
# the Graphviz package (http://graphviz.org/) in order to produce a
# pdf file with the corresponding reaction graph.
#
# For readability:
# * only the first few characters of a fraglet are shown;
# * "matchp" rules (which are conserved during reactions) are
#   represented only as reactions, not as species, except when they
#   are dynamically produced (in this case, a dotted arrow points from
#   the product to the corresponding reaction which it "catalyses")


use POSIX qw(strftime);

$inputf = shift(@ARGV);

sub initgraph {
    $molcnt = 0;
    $rulcnt = 0;
    $lnkcnt = 0;
}


sub addmol {
    local($s) = @_;
    if ($moltable{$s} eq "") {
	$molcnt++;
	$newmol = "n$molcnt";
	$moltable{$s} = $newmol;
	$prod{$newmol} = 0;
	$cons{$newmol} = 0;
	$indeg{$newmol} = 0;
	$outdeg{$newmol} = 0;
    }
    return $moltable{$s};
}


# addrule:
#   s = full rule, e.g. s="001 n002 -> n001 n003"
#   cmd = short label for rule, e.g. cmd="matchp:a", cmd="split"
#
sub addrule {
    local($s, $cmd) = @_;
    #print "ADDRULE \"$s\" \"$cmd\"\n";
    $r = $rultable{$s};
    if ($r eq "") {
	$rulcnt++;
	if ($cmd =~ /^matchp[ :\t]/ ) {
            $r = "R$rulcnt";
	} else {
	    $r = "r$rulcnt";
	}
	$rultable{$s} = $r;
	$cmdtable{$s} = $cmd;
    }
    return $r;
}


# addlnk: returns 1 if new link added, 0 if link was already there
sub addlnk {
    local($s) = @_;
    if ($lnktable{$s} eq "") {
	$lnkcnt++;
	$lnktable{$s} = $lnkcnt;
	return 1;
    }
    return 0;
}


sub printheader {
    $now = strftime "%Y-%m-%d@%H:%M", gmtime;
    print "strict digraph fraglet_pathway \{\n";
    print "  label=\"Fraglet pathway graph from $inputf, $now\";\n";
    print "  fontsize=8;\n";
    print "  fontname=Helvetica;\n";
    print "  edge [dir=forward,color=red];\n";
    print "  node [fontsize=10,fontname=Helvetica];\n";
}


sub printgraph {
    &printheader();

    print  "\n";
    printf "/* Full Molecule Table. Number of entries: %d\n", $molcnt;
    
    foreach $mol (keys %moltable) {
	$id  = $moltable{$mol};
	$in  = $indeg{$id};
	$out = $outdeg{$id};
	$cs  = $cons{$id};
	$pr  = $prod{$id};
	printf "  nid=%-4s in=%-2d out=%-2d prod=%-2d cons=%-2d mol=\"%s\" \n",
	$id, $in, $out, $pr, $cs, $mol;
    }
    print  "*/\n";

    print  "\n";
    print  "/* Active Molecules (graph nodes) */\n";
    print  "  node [shape=plaintext];\n";
    
    foreach $mol (keys %moltable) {
	#only print molecules that are linked to others
	$degree = $indeg{$moltable{$mol}} + $outdeg{$moltable{$mol}};
	if ($degree > 0) {
	    $r = $mol;
	    # strip lengthy molecules
	    if (length($r) > 12) {
		$r = substr($mol, 0, 12); $r = "${r}...";
	    }
	    printf "  {node [label=\"%s\"] %s}\n", $r, $moltable{$mol};
	}
    }
    
    print  "\n";
    printf "  /* Rule table. Number of entries: %d */\n", $rulcnt;
    print  "  node [shape=box];\n";

    foreach $i (keys %rultable) {
	$r = $rultable{$i};
	if ($r =~ /^R/) {
	    printf "  {node [style=filled,label=\"%s\"] %s} /* %s */\n",
	    $cmdtable{$i}, $r, $i;
	} else {
	    printf "  {node [label=\"%s\"] %s} /* %s */\n",
	    $cmdtable{$i}, $r, $i;
	}
    }
    
    printf "\n";
    print  "  /* Edges: */\n";
    foreach $i (keys %lnktable) {
	print "  $i;\n";
    }

    print "\}\n"
}

sub cleanfraglet {
    local($s) = @_;
    # remove leading and trailing blanks;
    $s =~ s/^[ \t]*//; $s =~ s/[ \t]*$//;
    # remove multiplicity of fraglet if any
    $s = $1 if ($s =~ /^([^\[]*\[[^\]]*\])[0-9]+/);
    # to do: remove blanks within fraglet
    return $s;
}

# make a short label for a long fraglet
sub mklabel {
    local($frl) = @_;
    $frl =~ s/^[^\[]*\[[ :\t]*//; # eliminate node name and bracket
    if ($frl =~ /^matchp?[ :\t]+/) {
        # match or matchp: keep second tag
	$frl =~ s/^(matchp?[ :\t]+[^ :\t]+)[ :\t]+.*$/$1/;
    } else {
	# label is first tag
	$frl =~ s/[ :\t]+.*$//;
    }
    return $frl;
}

&initgraph();

while (<>) {
    if (/^[0-9]*[ \t]*(.*)-->(.*)$/) {
	$r = $1;
	$p =$2;
	@ra = split(/,/, $r);
        @pa = split(/,/, $p);
	$react = "";

	# make nodes from reaction and its molecules
	for ($i=0; $i <= $#ra; $i++) {
	    $ra[$i] = cleanfraglet($ra[$i]);
	    $id = addmol($ra[$i]);
	    $cons{$id} ++;
	    $react="$react$id ";
	}
	$react="${react}->"; $i=0;
	for ($i=0; $i <= $#pa; $i++) {
	    $pa[$i] = cleanfraglet($pa[$i]);
	    $id = addmol($pa[$i]);
	    $prod{$id} ++;
	    $react="$react $id";
	}
	# make reaction node label from first reactant
	$rlabel = mklabel($ra[0]);
	$rule = addrule($react, $rlabel);

	# make links from reaction
	for ($i=0; $i <= $#ra && $i <= $#pa; $i++) {
	    $mol1 = $moltable{$ra[$i]};
	    $mol2 = $moltable{$pa[$i]};
	    if ($mol1 ne $mol2) {
		# *, b, * -> *, d, *
		if (addlnk("$mol1 -> $rule -> $mol2")) {
		    $outdeg{$mol1} ++;
		    $indeg{$mol2} ++;
		}
	    } elsif ($indeg{$mol1} > 0) {
		# enzyme that has been dynamically produced:
		# represented by a dotted arrow to reaction
		$outdeg{$mol1} ++ if (addlnk("$mol1 -> $rule [style=dotted]"));
	    }
	}
	while ($i <= $#ra) {
	    # a, b -> c
	    $mol = $moltable{$ra[$i++]};
	    $outdeg{$mol} ++ if (addlnk("$mol -> $rule"));
	}
	while ($i <= $#pa) {
	    # a -> b ,c
	    $mol = $moltable{$pa[$i++]};
	    $indeg{$mol} ++ if (addlnk("$rule -> $mol"));
	}
    }
}

&printgraph();

