/* fraglets0.29/src/frag.c

  fraglet interpreter

  Copyright (C) 2003-2007 <christian.tschudin@unibas.ch>

  This file is distributed under the terms of the GNU general Public
  License (GPL), see the file COPYING

  fraglet interpreter implementation

  History:

  20070121  christian.tschudin@unibas.ch: added 'x' command to exit
            without executing

  20070114  lidia.yamamoto@unibas.ch: bug fix: node_addsegment fixed so
            that it does not reinsert existing segments in node.

  20070112  lidia.yamamoto@unibas.ch:
    * added "length" transformation that returns the length of a fraglet:
      [length t1 a b c d] --> [t1 4 a b c d]

  20070111  lidia.yamamoto@unibas.ch:
    * added "sum" transformation [sum t1 2 3 tail] --> [t1 5 tail]
    * added "newname" transformation:
      [newname t1 prefix suffix tail] --> [t1 prefixsuffix tail]
    * added circular buffer as a palliative for name_lookup static buf
      problem for numbers
    * bug fix: waiting list was not being deleted by f_cleanup, so
      waiting fraglets would jump from one execution to the next.
    * bug fix in seg_parse: "a a", or "a[something]" would lead to
      Segmentation fault

  20070110  lidia.yamamoto@unibas.ch:
    * print production side of transformation when wait/delay expires:
      ... --> [fraglet that was waiting]
      this is important when calculating the concentration of fraglets
    * comments now accepted at the end of "a " and "f " lines;
    * removed the need of "e" at EOF (i.e. consider EOF='e');
    * added dynamic node creation:
      [newnode <seg> <node> tail] --> a <node> <seg>, <node>[tail]

  20070104  lidia.yamamoto@unibas.ch and christian.tschudin@unibas.ch:
	    added "delay", "copy" and "mutate" commands
  20061214  christian.tschudin@unibas.ch: removed need for ":" field sep,
            added the "empty" transformation
  20061213  christian.tschudin@unibas.ch: added number support, and
            the "lt" transformation
  20050526  Jasenko Zivanov (Unibas): added socket support
  20050220  lidia.yamamoto@unibas.ch: reset random seed for repeatability
  20050217  christian.tschudin@unibas.ch (fraglets v0.18):
    * fixed bug that didn't correctly take multiplicity into account for
      the probability of choosing a fraglet for reaction
    * improved debug and execution traces
  20050206  by author and lidia.yamamoto@unibas.ch (fraglets v0.17):
    * added (or recovered?) support for 'nul' operator
    * support for markers /m markername/ in genotype (selfman'05 and wac'05)
    * improved trace printout
  20040704  updated by christian.tschudin@unibas.ch for wac'04 (fraglets v0.15)
  200307??  created by christian.tschudin@unibas.ch (ains'03)
*/

#include "sockin.h"

struct frag *fraglist;
struct node *nodelist;
struct name *namelist;
int linecnt;

int debug;
//	swap comments to disable logging at all levels
#define DEBUG(l, stmt) if (l <= debug) do { stmt } while (0)
//#define DEBUG(l, stmt) 
int dump;

// J. Zivanov: loss emulation
int packet_loss_probability = 0;

extern char *name_lookup(symbol s);

static void f_print(FILE *fout, struct frag *f, int eol);

// ------------------------------------------------------------

// get symbol at position n in fraglet f
#define f_getsymbol(f,n)   ((f)->buf->data[(f)->offs + (n)])

// set symbol at position n to s, in fraglet f
#define f_setsymbol(f,n,s) ((f)->buf->data[(f)->offs + (n)] = (s))

// integer to symbol conversion
#define int2symbol(val) (SYM_NUMBEROFFS + (val))

int istrans(symbol s)
{
	return s > 0 && s <= SYM_LAST_TRANS;
}

int isnumber(symbol s)
{
	return s >= SYM_NUMBEROFFS;
}

unsigned short numberval(symbol s)
{
	return (s < SYM_NUMBEROFFS) ? 0 : s - SYM_NUMBEROFFS;
}

/*
int addhint(struct node *n, symbol s)
{
	DEBUG(9, printf("## addhint on node %s: %s\n", name_lookup(n->name), name_lookup(s)););
	if (n->hintcnt >= MAXHINTS) {
		DEBUG(10, fprintf(stderr, "MAXHINTS (%d) exceeded\n", MAXHINTS););
		return -1;
	}
	n->hints[n->hintcnt] = s;
	n->hintcnt++;

	return 0;
}
*/

// ------------------------------------------------------------

symbol name_add(char *cp)
{
	struct name *n = namelist;
	symbol i = 1;

	if ( isdigit(*cp) ) {
		int val;
		val = atoi(cp);
		if (val < 0)
			val = 0;
		return SYM_NUMBEROFFS + val;
	}

	while (n && strcmp(n->cp, cp)) {
		n = n->next;
		i++;
	}
	if (!n) {
		n = calloc(1, sizeof(*n));
		n->next = namelist;
		namelist = n;
		n->cp = malloc(strlen(cp) + 1);
		strcpy(n->cp, cp);
		n->sym = i;
	}
	
	return n->sym;
}


/* (lidia)
   warning!!! name_lookup returns static buffer for numbers:

   originally only one buffer position was available, so this wouldn't
   work: printf("%s, %s",name_lookup(number1), name_lookup(number2));

   I modified it to use 10 buffer positions in a circular way, so that
   up to 10 calls are supported without overwriting the buffer (this
   is a poor man's garbage collection...)
*/
char *name_lookup(symbol s)
{
	struct name *n;
	static int bidx = 0;
	static char buf[10][10];

	if (s >= SYM_NUMBEROFFS) {
	        bidx = (bidx+1)%10;
		sprintf(buf[bidx], "%d", s - SYM_NUMBEROFFS);
		return buf[bidx];
	}

	for (n = namelist; n; n = n->next)
		if (n->sym == s)
			return n->cp;
	return 0;
}


// ------------------------------------------------------------

struct node* node_lookup(symbol node, int create)
{
	struct node *n = nodelist;

	while (n && n->name != node)
		n = n->next;
	if (n)
		return n;
	if (!create)
		return 0;

	n = calloc(1, sizeof(*n));
	n->name = node;
	n->udp_wrapper = 0;
	n->location = Local;
	n->next = nodelist;
	nodelist = n;
	return n;
}

void node_addsegment(struct node *n, symbol segment)
{
	int i;

	if (!n) return;
	DEBUG(9, printf("## addtosegment %s %s\n",
			name_lookup(n->name), name_lookup(segment)););

	// (lidia): do not reinsert existing segment
	for (i = 0; i < MAXSEGMENTS && n->segments[i] &&
	       n->segments[i] != segment; i++);
	if (i < MAXSEGMENTS) {
	        // (superfluous): if (n->segments[i] == segment) return;
		n->segments[i] = segment;
	}
	else { // complain ...
	        fprintf(stderr, "no segment place for seg %s in node %s\n",
			name_lookup(segment), name_lookup(n->name));
		exit(-1);
	}
}

// (lidia) segment lookup: true if segment found in node
//
int seg_lookup(struct node *n, symbol seg)
{
	int i;

	if (!n) return 0;
	for (i = 0; i < MAXSEGMENTS && n->segments[i]; i++)
		if (n->segments[i] == seg)
			return 1;
	return 0;

}


// ------------------------------------------------------------

/* 
struct fragbuf* b_new(int len) // len in bits
{
	struct fragbuf *b;

	b = (struct fragbuf*) malloc(sizeof(struct fragbuf) + (len + 7)/8);
	if (b) {
		b->refcnt = b->len = 0;
	}
	return b;
}
*/

void b_free(struct fragbuf *buf)
{
	free(buf);
}

// ------------------------------------------------------------

struct frag* f_new(struct fragbuf *b, int offs, int len, int mult, symbol n)
{
	struct frag *f = (struct frag*) calloc(1, sizeof(*f));
	if (!f)
		return 0;
	// should check offs and len...
	f->buf = b;
	f->next = 0;
	f->nd = n;
	f->offs = offs;
	f->len = len;
	f->mult = mult;
	b->refcnt++;
	return f;
}

symbol f_peek(struct frag *f)
{
	if (f->offs + 1 > f->buf->len)
		return 0;
	else {
		return f->buf->data[f->offs];
/*	symbol s;
	memcpy(&s, f->buf->data + f->offs, sizeof(s));
	return s;
*/
	}
}


void f_free(struct frag *f)
{
	f->buf->refcnt--;
	if (f->buf->refcnt <= 0)
		b_free(f->buf);
	free(f);
}


void f_freeLinkedList(struct t_queue *q)
{
	struct frag *f;

	while (q->first) {
		f = q->first;
		q->first = f->next;
		f_free(f);
	}
	q->last = 0;
}


int ismatch(symbol s)
{
	return s == SYM_MATCH || s == SYM_MATCHP;
}


void t_append(struct frag *f)
{
	struct node *n = node_lookup(f->nd, 0);
	symbol h = f->buf->data[f->offs];

	DEBUG(9, printf("# append transformation %s\n", name_lookup(h)););

	f->next = 0;
	if (n->tq.last)
		n->tq.last->next = f;
	else
		n->tq.first = f;
	n->tq.last = f;
	//	addhint(n, h);
}


void w_append(struct frag *f, int duration)
{
	struct node *n = node_lookup(f->nd, 0);

	DEBUG(9, printf("# append wait %s %d\n",
			name_lookup(f->buf->data[f->offs]), duration););

	f->next = 0;

	// pending: modify wait_duration to real time instead of virtual

	f->expires = n->opcnt + duration;
	if (n->wait.last) {
		struct frag *p;

		if (n->wait.first->expires > f->expires) {
			f->next = n->wait.first;
			n->wait.first = f;
		} else {
			for (p = n->wait.first; p != n->wait.last; p = p->next) {
				if (p->next->expires > f->expires) {
					f->next = p->next;
					p->next = f;
					break;
				}
			}
			if (p == n->wait.last) {
				n->wait.last->next = f;
				n->wait.last = f;
			}
		}
	} else {
		n->wait.first = f;
		n->wait.last = f;
	}
}


void m_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'match' for symbol %s\n", name_lookup(h)););

	// (lidia): also accept numbers as match tags
	if (isnumber(h)) h = SYM_NUMBER;
	if (h >= MAXSYMBOLS) {
	  printf("m_add: Array out of bound exception: mtab[%d]\n",h);
	  exit(-1);
	}
	DEBUG(9, printf("## maxsym=%d, sym_noff=%d, h=%d\n",MAXSYMBOLS,SYM_NUMBEROFFS,h););

	f->next = 0;
	if (n->mtab[h].last)
		n->mtab[h].last->next = f;
	else
		n->mtab[h].first = f;
	n->mtab[h].last = f;
/*
	if (n->otab[h].first)
		addhint(n, h);
*/
}


void o_add(struct node *n, symbol h, struct frag *f)
{
	DEBUG(9, printf("# adding 'other' for symbol %s\n", name_lookup(h)););

	// (lidia): also accept numbers as front symbols
	if (isnumber(h)) h = SYM_NUMBER;
	if (h >= MAXSYMBOLS) {
	  printf("o_add: Array out of bound exception: otab[%d]\n",h);
	  exit(-1);
	}
	DEBUG(9, printf("## maxsym=%d, sym_noff=%d, h=%d\n",MAXSYMBOLS,SYM_NUMBEROFFS,h););

	f->next = 0;

	if (n->otab[h].last)
		n->otab[h].last->next = f;
	else
		n->otab[h].first = f;
	n->otab[h].last = f;
	/*
	if (n->mtab[h].first)
		addhint(n, h);
	*/
}


struct frag* rx_fraglet(struct frag *f) // , int offs, int len)
{
	symbol head;
	struct node *n;

	if (!f) return 0;

	// J. Zivanov: moved the following line here:
	n = node_lookup(f->nd, 1);	

	DEBUG(9, printf("## rx_fraglet\n"););

	if (!n || !f) return 0;
	// ...from here

	if (f->len == 0) {
		f_free(f);
		return 0;
	}
	/*
	{
		int i;
		for (i = 0; i < f->len; i++) {
			if (*name_lookup(f->buf->data[f->offs+i]) == 'p') {
				score++;
				break;
			}
		}
	}
	*/
	head = f_peek(f);
	if (istrans(head)) {
		t_append(f);
	} else {
		if (ismatch(head)) {
			DEBUG(9, printf("## match\n"););
			if (f->len == 1) {
				f_free(f);
				return 0;
			}
			head = f->buf->data[f->offs + 1];
			m_add(n, head, f);
		} else
			o_add(n, head, f);
	}
	n->possibleHint |= 1;
	return f;
}


struct frag* f_parsefrag(int lno, char *line)
{
	char *cp, *a, *b;
	int m;
	struct fragbuf *buf;
	symbol n;

	// format: "poolname [ sym1 : sym2 : ... ] mult"

	a = strchr(line, '[');
	b = strchr(line, ']');
	if (!a || !b) {
		// this also captures markers of the form:  /m markername/
		return 0;
	}
	*a = *b = 0;

	cp = strtok(line, " \n\t\v");
	//if (!cp) {
	  // create default node with empty name
	  // fprintf(stderr, "parse error line %d --- no node prexif\n", lno);
	  //return 0;
	//}
	n = name_add(line);

	cp = strtok(b+1, " \n\t\v");
	if (!cp || *cp == '#') // default multiplicity is 1
		m = 1;
	else
		m = atoi(cp);

	if (m <= 0) {
		fprintf(stderr, "parse error line %d --- multiplicity must be > 0\n", lno);
		return 0;
	}

	buf = calloc(1, sizeof(*buf));
	for (a = strtok(a+1, " :\t\n\v"); a; a = strtok(NULL, " :\t\n\v")) {
		buf->len++;
		buf = realloc(buf, sizeof(*buf) + buf->len * sizeof(symbol));
		buf->data[buf->len - 1] = name_add(a);
	}

	return f_new(buf, 0, buf->len, m, n);
}


// parse genotype
struct frag* f_parse(int lno, char *line)
{
	char *t;
	struct frag *flist = 0, *f;

	if (*line != '/')
		return 0;

	for (t = strtok(line+1, "/\n"); t ; t = strtok(NULL, "/\n")) {
		f = f_parsefrag(lno, t);
		if (f) {
			f->next = flist;
			flist = f;
		}
	}
	return flist;
}


symbol seg_parse(int lno, char *line, symbol *node, char* ip_address, int* port)
{
	char *cp;
	symbol s;

	// J. Zivanov: new segment format (UDP support):
	// "A <nodename> <segname> [(<IPv4 address>|"local") <port>]"

	if (strchr(line, '[') || strchr(line, ']')) {

	  // (lidia): if [ or ] in node/seg descr then they cannot be
	  // used in fraglets, so reject this line (probably this is a
	  // fraglet and the user forgot to start the line with 'f')

	        return 0;
	}

	cp = strtok(line+1, " \t\n");
	if (!cp || *cp == '#') {
		fprintf(stderr, "syntax error on line %d: missing node name\n", lno);
		return 0;
	}

	*node = name_add(cp);

	cp = strtok(0, " \t\n");
	if (!cp || *cp == '#') {
		fprintf(stderr, "syntax error on line %d: missing segment name\n", lno);
		return 0;
	}
	s = name_add(cp);

	cp = strtok(0, " \t\n");
	if (cp && *cp != '#')
	{
	  // UDP segment
		if (strcmp(cp, "local"))
		{
			memccpy(ip_address, cp, '\n', 15);
			ip_address[15] = '\n';
		}
		else ip_address[0] = 'l';
		cp = strtok(0, " \t\n");
		*port = atoi(cp);
		cp = strtok(0, " \t\n");
	}
	else ip_address[0] = '\0';
	
	if (!*node || !s || cp && *cp != '#') {
		fprintf(stderr, "syntax error on line %d\n", lno);
		return 0;
	}
	return s;
}


int f_import(FILE *fin)
{
	static int iseof = 0;

	char line[10000], *cp;
	struct frag *f, *f1, *f2;
	symbol node, seg;

	// J. Zivanov: udp support
	char ip_address[32];
	int port;

	if (iseof) return -1;
	f1 = NULL;
	while (fgets(line, sizeof(line)-1, fin)) {
		linecnt++;
		switch(line[0]) {
		case '\n':
		case '#':
			continue;
		case 'e':
			return 0;
		case 'f':
		case 'F':
			for (cp = line+1; isspace(*cp); cp++);
			f = f_parsefrag(linecnt, cp);
			if (!f) {
			  // Lidia: 'line' is broken at this stage, cannot be printed
				fprintf(stderr, "Line %d: cannot parse fraglet\n", linecnt);
				exit(-1);
			} else f1 = f;
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			continue; //break;
		case 'g':
		case 'G':
			for (cp = line+1; isspace(*cp); cp++);
			f = f_parse(linecnt, cp);
			if (!f) {
				fprintf(stderr, "Line %d: cannot parse genotype\n", linecnt);
				exit(-1);
			}
			while (f) {
				f2 = f->next;
				f->next = 0;
				rx_fraglet(f);
				f = f2;
			}
			return 0;
		case 'a':
		case 'A':
		  // J. Zivanov: added udp support for segment
			seg = seg_parse(linecnt, line, &node, ip_address, &port);
			if (!seg) {
				fprintf(stderr, "Line %d: cannot parse segment %s\n", linecnt, line);
				exit(-1);
			}
			node_addsegment(node_lookup(node, 1), seg);
			if (ip_address[0] != '\0') 
			{	
				if (ip_address[0] == 'l') open_local_socket(node, port);
				else open_socket_to_remote_node(node, ip_address, port);
			}
			continue; // break;
		case 'm':
		case 'M':
			continue; // break;
		case 'x':
			return -1;
		default:
			fprintf(stderr, "Line %d: cannot parse %s\n", linecnt, line);
			exit(-1);
		}
	}
	iseof = 1;
	if (f1) return 0;
	else return -1;
}


void f_compact_list(struct frag *f, struct frag **last)
{
	struct frag *f2, *f3, **fpp;

	while (f) {
		int len = f->len;
		symbol *sp = f->buf->data + f->offs;

		for (fpp = &f->next, f2 = f->next; f2; ) {
			if (len != f2->len ||
			    memcmp(sp, f2->buf->data + f2->offs,
				   sizeof(symbol) * len)) {
				fpp = &f2->next;
				f2 = f2->next;
				continue;
			}
			// same content!
			f->mult += f2->mult;
			f3 = f2->next;
			f_free(f2);
			f2 = *fpp = f3;
		}
		if (!f->next)
			*last = f;
		f = f->next;
	}
}


static void f_print(FILE *fout, struct frag *f, int eol)
{
	int i;

	fprintf(fout, "%s", name_lookup(f->nd));

	for (i = 0; i < f->len; i++) {
		fprintf(fout, i ? " %s" : "[%s",
		       name_lookup(f->buf->data[f->offs + i]));
	}

	fprintf(fout, eol ? "]\n" : "]");
}


static void f_export_list(FILE *fout, struct node *n, struct frag *f)
{
	for (; f; f = f->next) {
		fprintf(fout, "f ");
		f_print(fout, f, 0);
		fprintf(fout, "%d\n", f->mult);
	}
}


static void f_export_mtab(FILE *fout, struct node *n, struct m_rec *mtab)
{
	int i;

	for (i = 0; i < MAXSYMBOLS; i++, mtab++) {
		f_compact_list(mtab->first, &(mtab->last));
		f_export_list(fout, n, mtab->first);
	}
}


int f_export(FILE *fout)
{
	struct node *n;

	for (n = nodelist; n; n = n->next) {
		int i;
		char *nname = name_lookup(n->name);
		if (n->segments[0]) {
			for (i = 0; n->segments[i]; i++) {
				fprintf(fout, "a %s %s\n", nname,
				       name_lookup(n->segments[i]));
			}
		}
	}

	for (n = nodelist; n; n = n->next) {
		f_compact_list(n->tq.first, &(n->tq.last));
		f_export_list(fout, n, n->tq.first);
		f_compact_list(n->wait.first, &(n->wait.last));
		f_export_list(fout, n, n->wait.first);
		f_export_mtab(fout, n, n->mtab);
		f_export_mtab(fout, n, (struct m_rec*) n->otab);

	}

	return 0;
}


/* f_shrink removes the given number of front symbols. If
   empty, 0 is returned. Otherwise a fraglet (copy) is returned
   with a multiplicity of 1. Note that this copy still might
   share the buffer with some other fraglet, so you should not
   modify the content.
*/
struct frag* f_shrink(struct frag *f, int cnt)
{
	DEBUG(9, printf("# f_shrink %d\n", f->mult););

	if (f->mult == 1) {
		if (f->len <= cnt) {
			f_free(f);
			return 0;
		}
		f->offs += cnt;
		f->len -= cnt;
	} else {
		f->mult--;
		rx_fraglet(f);
		if (f->len > cnt)
			f = f_new(f->buf, f->offs + cnt,
				  f->len - cnt, 1, f->nd);
	}

	return f;
}


// ----------------------------------------------------------------------

struct frag* f_nop(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	// Lidia: added execution trace
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	f = f_shrink(f, 1);

	if (f) {
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else
		DEBUG(1, printf("[]\n"););

	return f;
}


// lidia: copy entire fraglet
struct frag* f_copy(struct frag *f)
{
	DEBUG(9, printf("# f_copy %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	if (f->len < 2)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 1;
	f = f_shrink(f, 1);
	if (f) {
		f->mult++;
		node_lookup(f->nd, 0)->symbalance -= f->len;
		DEBUG(1, f_print(stdout, f, 0););
		DEBUG(1, printf("%d\n", f->mult););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_delay(struct frag *f)
{
	DEBUG(9, printf("# f_delay %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> ...\n"););
	      
	if (f->len < 3)
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 1]))
		goto drop;

	f = f_shrink(f, 0);
	if (f)
		w_append(f, numberval(f->buf->data[f->offs + 1]));

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;

}


struct frag* f_dup(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_dup %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	if (f->len < 3)
		goto drop;

// dup is symbol count neutral:
// 	node_lookup(f->nd, 0)->symbalance += 0;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->buf->data[f->offs] = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * f->len);
	buf->len = f->len;
	buf->refcnt = 0;
	memcpy(buf->data + 2, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);

	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_empty(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;
	symbol s;

	DEBUG(9, printf("# f_empty %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->len > 3)
		s = f->buf->data[f->offs + 2];
	else
		s = f->buf->data[f->offs + 1];

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
		f->buf->data[f->offs] = s;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data+1, f->buf->data + f->offs + 3,
	       sizeof(symbol) * (f->len - 3));
	buf->data[0] = s;
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_exch(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_exch %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 4)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 1;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		f->offs++;
		f->len--;
		s = f->buf->data[f->offs + 1];
		f->buf->data[f->offs + 1] = f->buf->data[f->offs + 2];
		f->buf->data[f->offs + 2] = s;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 1));
	buf->len = f->len - 1;
	buf->refcnt = 0;
	memcpy(buf->data + 3, f->buf->data + f->offs + 4,
	       sizeof(symbol) * (f->len - 4));
	buf->data[0] = f->buf->data[f->offs + 1];
	buf->data[1] = f->buf->data[f->offs + 3];
	buf->data[2] = f->buf->data[f->offs + 2];
	f2 = f_new(buf, 0, f->len - 1, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_mutate(struct frag *f)
{
	int num, pos;
	symbol s;

	DEBUG(9, printf("# f_mutate %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 3)
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 1]))
		goto drop;
	num = numberval(f->buf->data[f->offs + 1]);

	node_lookup(f->nd, 0)->symbalance += 2;
	f = f_shrink(f, 2);
	if (!f) {
		DEBUG(1, printf("[]\n"););
		return 0;
	}
	if (num == 0 || f->len < 3 || ((rand() % num) != 0)) {
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	if (f->buf->refcnt != 1) { // create our own copy
		struct fragbuf *buf;
		struct frag *f2;
		buf = malloc(sizeof(*buf) + sizeof(symbol) * f->len);
		buf->len = f->len;
		buf->refcnt = 0;
		memcpy(buf->data, f->buf->data + f->offs,
		       sizeof(symbol) * f->len);
		f2 = f_new(buf, 0, f->len, 1, f->nd);
		f_free(f);
		f = f2;
	}

	pos = rand() % (f->len-2);
	s = f->buf->data[f->offs + 1 + pos];
	f->buf->data[f->offs + 1 + pos] = f->buf->data[f->offs + 2 + pos];
	f->buf->data[f->offs + 2 + pos] = s;
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


/* (lidia): copy a fraglet buffer (nasty f_new() doesn't do it!!!)
  such that it can be safely updated (after f_shrink does its dirty
  job); ignores already consumed headers, so buffer might be smaller
  than the original one. when the fraglet is already standalone
  (multiplicity is one and no one else is pointing to its buffer
  (refcnt=1) then fraglet is not changed.
*/
struct frag *f_bufcpy(struct frag *f) {
        struct fragbuf *newbuf;

	if (f->mult == 1 && f->buf->refcnt == 1) return f;
	if (f->mult > 1) f = f_shrink(f, 0);

	newbuf = malloc(sizeof(struct fragbuf) + sizeof(symbol) * f->len);
	newbuf->len = f->len;
	newbuf->refcnt = 1;
	memcpy(newbuf->data, f->buf->data + f->offs, sizeof(symbol) * f->len);
	
	f->buf->refcnt--;
	if (f->buf->refcnt <= 0) b_free(f->buf);
	f->buf = newbuf;
	f->offs = 0;
	return f;
}


// (lidia)
// 
// [length tag <tail>] --> [tag <taillength> <tail>]
//
struct frag* f_length(struct frag *f)
{
        symbol tag, tlen;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 2)
		goto drop;
	
	tag = f_getsymbol(f,1);

	tlen = int2symbol(f->len - 2);

        f = f_bufcpy(f);
	f_setsymbol(f,0,tag);
	f_setsymbol(f,1,tlen);
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}



// (lidia)
// [sum tag n1 n2 tail] --> [tag (n1+n2) tail]
//
struct frag* f_sum(struct frag *f)
{
	symbol tag, n1, n2, ns;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;
	
	tag = f_getsymbol(f,1);
	n1  = f_getsymbol(f,2);
	n2  = f_getsymbol(f,3);
	if (!isnumber(n1) || !isnumber(n2)) goto drop;
	ns = int2symbol(numberval(n1) + numberval(n2));

	f = f_shrink(f,2);
	if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,tag);
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// (lidia)
// [diff tag n1 n2 tail] --> [tag abs(n1-n2) tail]
//
struct frag* f_diff(struct frag *f)
{
	symbol tag, n1, n2, ns;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;
	
	tag = f_getsymbol(f,1);
	n1  = f_getsymbol(f,2);
	n2  = f_getsymbol(f,3);
	if (!isnumber(n1) || !isnumber(n2)) goto drop;
	ns = int2symbol(abs(numberval(n1) - numberval(n2)));

	f = f_shrink(f,2);
	if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,tag);
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


// (lidia)
// [newname tag s1 s2 tail] --> [tag s1s2 tail]
//
struct frag* f_newname(struct frag *f)
{
	symbol tag, s1, s2, ns;
	char *str1, *str2, *newstr;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;
	
	tag = f_getsymbol(f,1);
	s1  = f_getsymbol(f,2);
	s2  = f_getsymbol(f,3);
	str1 = name_lookup(s1);
	str2 = name_lookup(s2);
	newstr = (char *) malloc(strlen(str1) + strlen(str2) + 1);
	strcpy(newstr,str1);
	strcat(newstr,str2);
	ns = name_add(newstr);
	free(newstr);

	f = f_shrink(f,2);
	if (f) {
	        f = f_bufcpy(f);
		f_setsymbol(f,0,tag);
		f_setsymbol(f,1,ns);
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}

	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}



struct frag* f_newnode(struct frag *f)
{
	symbol seg;
	struct node *newnode;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 3)
		goto drop;
	
	seg = f_getsymbol(f,1);
	
	newnode = node_lookup(f_getsymbol(f,2), 1);
	if (!newnode)
		goto drop;

	DEBUG(9, printf("## newnode %s seg=%s (%d)\n",
			name_lookup(newnode->name), name_lookup(seg),
			f->mult););

	// (superfluous): if (!seg_lookup(newnode, seg))
	node_addsegment(newnode, seg);
	f = f_shrink(f,3);
	if (f) {
		f->nd = newnode->name;
		node_lookup(f->nd, 0)->symbalance -= f->len;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
	} else {
		DEBUG(1, printf("[]\n"););
	}
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_nul(struct frag *f)
{
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_pop(struct frag *f)
{
	DEBUG(9, printf("# f_pop %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 3)
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		f->offs += 2;
		f->len -= 2;
		f->buf->data[f->offs] = f->buf->data[f->offs - 1];
	} else {
		struct fragbuf *b2;
		struct frag *f2;

		b2 = malloc(sizeof(*b2) + sizeof(symbol) * (f->len - 2));
		b2->len = f->len - 2;
		b2->refcnt = 0;
		memcpy(b2->data + 1, f->buf->data + f->offs + 3,
		       sizeof(symbol) * (f->len - 3));
		b2->data[0] = f->buf->data[f->offs + 1];
		f2 = f_new(b2, 0, f->len - 2, 1, f->nd);
		if (--(f->mult) <= 0)
			f_free(f);
		else {
			t_append(f);
//			addhint(node_lookup(f->nd, 0), SYM_POP);
		}
		f = f2;
	}
	DEBUG(1, f_print(stdout, f, 1););
	rx_fraglet(f);
	return f;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) <= 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_split(struct frag *f)
{
	symbol *sp;
	int i;
	struct frag *f2;

	DEBUG(9, printf("# f_split %d\n", f->mult););
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	for (i = 1, sp = f->buf->data + f->offs + 1; i < f->len; i++, sp++) {
		if (*sp == SYM_STAR)
			break;
	}
	if (i == 1) { // empty first half
		node_lookup(f->nd, 0)->symbalance += 2;
		f = f_shrink(f, 2);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
			
		rx_fraglet(f);
		return f;
	}
	if (i >= f->len) { // no star (empty second half)
		node_lookup(f->nd, 0)->symbalance += 1;
		f = f_shrink(f, 1);
		DEBUG(1, if (f) f_print(stdout, f, 1); else printf("[]\n"););
		rx_fraglet(f);
		return f;
	}
	node_lookup(f->nd, 0)->symbalance += 2;
	if (i != (f->len - 1)) { // accept the second half, if not empty
		f2 = f_new(f->buf, f->offs + i + 1, f->len - i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 0); printf(", "););
		rx_fraglet(f2);
	}
	// handle the first half
	if (f->mult == 1) {
		f->offs += 1;
		f->len = i - 1;
		DEBUG(1, f_print(stdout, f, 1););
	} else {
		f->mult--;
		f2 = f_new(f->buf, f->offs + 1, i - 1, 1, f->nd);
		DEBUG(1, f_print(stdout, f2, 1););
		rx_fraglet(f2);
	}
	rx_fraglet(f);
	return 0;
}


// J. Zivanov: changed this to send to socket when specified in topology
// line of destination; must also check whether topology line of
// sender has a compatible port & addr combination, and whether they
// share the same "segment" ; actually we can specify that all nodes
// able to communicate via udp share the same segment - we could even
// call this segment just "udp".

struct frag* f_send(struct frag *f)
{
	symbol seg;
	struct node *dest, *n;
	struct frag *f2;
	int i;

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););

 	if (f->len < 4)
		goto drop;
	
	seg = f->buf->data[f->offs + 1];
	
	dest = node_lookup(f->buf->data[f->offs + 2], 0);
	if (!dest)
		goto drop;

	DEBUG(9, printf("## send from %s to %s (%d)\n",
			name_lookup(f->nd), name_lookup(dest->name),
			f->mult););

	n = node_lookup(f->nd, 0);

	/* (lidia): safe to discard this? (replaced by seg_lookup)
	for (i = 0; i < MAXSEGMENTS && n->segments[i]; i++)
		if (n->segments[i] == seg)
			break;
	if (i >= MAXSEGMENTS || !n->segments[i])
		goto drop;

	for (i = 0; i < MAXSEGMENTS && dest->segments[i]; i++)
		if (dest->segments[i] == seg)
			break;
	if (i >= MAXSEGMENTS || !dest->segments[i])
		goto drop;
	*/
	if (!seg_lookup(n,seg) || !seg_lookup(dest,seg)) goto drop;

	// J. Zivanov
	switch (dest->location)
	{
		case Local:
		
			node_lookup(f->nd, 0)->symbalance -= f->len + 3;
			node_lookup(dest->name, 0)->symbalance += f->len - 3;
		
			if (f->mult == 1) {
				f->offs += 3;
				f->len -= 3;
				f2 = f;
				f2->nd = dest->name;
			} else {
				f->mult--;
				f2 = f_new(f->buf, f->offs + 3, f->len - 3, 1, dest->name);
				t_append(f);
		//		addhint(node_lookup(f->nd, 0), SYM_SEND);
			}
			DEBUG(1, f_print(stdout, f2, 1););
			rx_fraglet(f2);
			
			break;
		
		case Remote:
		
			node_lookup(f->nd, 0)->symbalance -= f->len + 3;
		
			if (f->mult == 1) {
				f->offs += 3;
				f->len -= 3;
				f2 = f;
				f2->nd = dest->name;
			} else {
				f->mult--;
				f2 = f_new(f->buf, f->offs + 3, f->len - 3, 1, dest->name);
				t_append(f);
		//		addhint(node_lookup(f->nd, 0), SYM_SEND);
			}
			DEBUG(1, f_print(stdout, f2, 0); printf(" (unreliably)\n"););
			if (!packet_loss_probability || rand() % packet_loss_probability)
				send_fraglet_through_socket(f2, n, dest);
			else {DEBUG(1, printf("Loss emulator: OUTBOUND PACKET LOST\n"););}
			
			break;
	}
	
	return f;
drop:
	node_lookup(f->nd, 0)->symbalance += f->len;	
	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}


struct frag* f_lt(struct frag *f)
{
	struct fragbuf *buf;
	struct frag *f2;

	DEBUG(9, printf("# f_lt %d\n", f->mult););

	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> "););
	      
	if (f->len < 5)
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 3]))
		goto drop;
	if (!isnumber(f->buf->data[f->offs + 4]))
		goto drop;

	node_lookup(f->nd, 0)->symbalance += 2;

	if (f->mult == 1 && f->buf->refcnt == 1) {
		symbol s;
		if (numberval(f->buf->data[f->offs + 3]) <
		    numberval(f->buf->data[f->offs + 4]))
			f->buf->data[f->offs + 2] = f->buf->data[f->offs + 1];
		f->offs += 2;
		f->len -= 2;
		DEBUG(1, f_print(stdout, f, 1););
		rx_fraglet(f);
		return 0;
	}

	buf = malloc(sizeof(*buf) + sizeof(symbol) * (f->len - 2));
	buf->len = f->len - 2;
	buf->refcnt = 0;
	memcpy(buf->data, f->buf->data + f->offs + 2,
	       sizeof(symbol) * (f->len - 2));
	if (numberval(f->buf->data[f->offs + 3]) <
	    numberval(f->buf->data[f->offs + 4]))
		buf->data[0] = f->buf->data[f->offs + 1];
	f2 = f_new(buf, 0, f->len - 2, 1, f->nd);
	DEBUG(1, f_print(stdout, f2, 1););
	rx_fraglet(f2);

	if (--(f->mult) <= 0)
		f_free(f);
	else
		rx_fraglet(f);
	return 0;

drop:
	node_lookup(f->nd, 0)->symbalance += f->len;

	DEBUG(1, printf("[]\n"););
	if (--(f->mult) == 0)
		f_free(f);
	else
		t_append(f);
	return 0;
}

struct frag* f_wait(struct frag *f)
{
	DEBUG(1, printf("%05d  ", node_lookup(f->nd, 0)->opcnt);
	      f_print(stdout, f, 0); printf(" --> ...\n"););
	      
	f = f_shrink(f, 0);
	if (f)
	  w_append(f, WAIT_DURATION);
	return f;
}


struct frag* f_pick(struct t_queue *q, int offs)
{
	struct frag *f, *prev;

/*
	printf("f_pick %d\n", offs);
	printf("\n* - - - - - - - - - - -\n");
	f_export(stdout);
	printf("\n* - - - - - - - - - - -\n");
*/


	// pick the fraglet at offset 'offs', taking its
	// multiplicity into account
	prev = 0;
	for (f = q->first; f; f = f->next) {
//		printf("f_pick mult=%d\n", f->mult);
		if (offs < f->mult)
			break;
		offs -= f->mult;
		prev = f;
	}
	if (!f) {
		fprintf(stderr, "# f_pick: no such fraglet\n");
		return 0;
	}
//	printf("f_pick found %d, next=%p\n", offs, f->next);

	// remove the fraglet from the queue
	if (!prev) {
		q->first = f->next;
	} else {
		prev->next = f->next;
	}
	if (q->last == f)
		q->last = prev;
	f->next = 0;

	return f;
}

int f_execFromQueue(struct node *n, struct t_queue *q, int offs)
{
	struct fragbuf *buf;
	struct frag *f, *prev, *f1, *f2, *f3;
	int ops = 0;
	int s;

	DEBUG(9, printf("## f_execFromQueue %d\n", offs););

	f = f_pick(q, offs);
	if (!f)
		return 0;

//	printf("f_exec mult=%d\n", f->mult);

	s = f_peek(f);
	if (istrans(s)) {
/*
		struct frag *f = n->tq.first;
		if (!f)
			return ops;
		n->tq.first = f->next;
		if (n->tq.last == f)
			n->tq.last = 0;
		f->next = 0;
		s = f_peek(f);
*/
		switch (s) {
		case SYM_NOP:	 f_nop(f);       break;
		case SYM_COPY:   f_copy(f);      break;
		case SYM_DELAY:  f_delay(f);	break;
		case SYM_DUP:    f_dup(f);       break;
		case SYM_EXCH:   f_exch(f);      break;
		case SYM_EMPTY:  f_empty(f);     break;
		case SYM_LT:	 f_lt(f);	break;
		case SYM_LENGTH: f_length(f);	break;
		case SYM_DIFF:	 f_diff(f);	break;
		case SYM_SUM:	 f_sum(f);	break;
		case SYM_NEWNAME:    f_newname(f);   break;
		case SYM_NEWNODE:    f_newnode(f);   break;
		case SYM_MUTATE: f_mutate(f);	break;
		case SYM_NUL:    f_nul(f);	break;
		case SYM_POP:	 f_pop(f);       break;
		case SYM_SEND:   f_send(f);      break;
		case SYM_SPLIT:  f_split(f);     break;
		case SYM_WAIT:   f_wait(f);	break;
		case SYM_TRIGGER: appl_trigger(f); break;
		default:
			fprintf(stderr, "'%s' not implemented\n",
			       name_lookup(s));
			if (--(f->mult) == 0)
				f_free(f);
			break;
		}
		return ops+1;
	}

//	printf("reaction ...\n");
	if (ismatch(s)) {
		struct o_rec *or;
		int i;
		symbol tag;
		f1 = f;

		// (lidia): match a number
		tag = f_getsymbol(f,1);
		if (isnumber(tag)) {
		  or = n->otab + SYM_NUMBER;
		  // (lidia): stopped here!
		} else {
		  or = n->otab + tag;
		}
		i = random() % or->active_cnt;
/*
		printf("found f1, now picking match %d from %d\n",
		       i, or->active_cnt);
*/
		f2 = f_pick((struct t_queue*) or, i);
		if (!f2) {
			rx_fraglet(f1);
			return 0;
		}
	} else {
		struct m_rec *mr;
		int i;

		f2 = f;
		mr = n->mtab + s;
		i = random() % mr->active_cnt;
/*
		printf("found f2, now picking match %d from %d\n",
		       i, mr->active_cnt);
*/
		f1 = f_pick((struct t_queue*) mr, i);
		if (!f1) {
			rx_fraglet(f2);
			return 0;
		}
	}

/*
	f1 = n->mtab[s].first;
	f2 = n->otab[s].first;
	if (!f1 || !f2)
		return ops;
*/
	
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f1, 0); printf(", ");
	      f_print(stdout, f2, 0); printf(" --> "););
	      
	if ((f1->len + f2->len - 3) > 0) {
		buf = malloc(sizeof(*buf) + sizeof(symbol) * (f1->len + f2->len - 3));
		buf->refcnt = 0;
		buf->len = f1->len + f2->len - 3;
		memcpy(buf->data, f1->buf->data + f1->offs + 2, sizeof(symbol)*(f1->len - 2));
		memcpy(buf->data + f1->len - 2, f2->buf->data + f2->offs + 1, sizeof(symbol)*(f2->len - 1));
		f3 = f_new(buf, 0, buf->len, 1, f1->nd);
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, f_print(stdout, f3, 1););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", ");
			      f_print(stdout, f3, 1););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
		node_lookup(f3->nd, 0)->symbalance -= f3->len;
		rx_fraglet(f3);
	} else {
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, fprintf(stdout, "[]\n"););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", []\n"););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
	}

/*
	n->mtab[s].first = f1->next;
	if (n->mtab[s].last == f1)
		n->mtab[s].last = 0;
	n->otab[s].first = f2->next;
	if (n->otab[s].last == f2)
		n->otab[s].last = 0;
*/

	if (f1->buf->data[f1->offs] == SYM_MATCH) {
		node_lookup(f1->nd, 0)->symbalance += f1->len;
		f1->mult--;
	} // else SYM_MATCHP
	if (f1->mult <= 0)
		f_free(f1);
	else
		rx_fraglet(f1);

	node_lookup(f2->nd, 0)->symbalance += f2->len;
	f2->mult--;
	if (f2->mult <= 0)
		f_free(f2);
	else
		rx_fraglet(f2);

/*
	printf("\n# - - - - - - - - - - -\n");
	f_export(stdout);
	printf("\n# - - - - - - - - - - -\n");
*/

	return ops+1;
	
}


// (lidia): shall we erase this?
#if 0
int f_exec(struct node *n, symbol s, int offs)
{
	struct fragbuf *buf;
	struct frag *f1, *f2, *f3;
	int ops = 0;

	DEBUG(9, printf("## f_exec %s\n", name_lookup(s)););

	if (istrans(s)) {
		struct frag *f = n->tq.first;
		if (!f)
			return ops;
		n->tq.first = f->next;
		if (n->tq.last == f)
			n->tq.last = 0;
		f->next = 0;
		s = f_peek(f);
		switch (s) {
		case SYM_NOP:   f_nop(f);       break;

		case SYM_DUP:   f_dup(f);       break;
		case SYM_EXCH:  f_exch(f);      break;
		case SYM_EMPTY: f_empty(f);     break;
		case SYM_NEW:   f_newname(f);   break;
		case SYM_NEWNODE:   f_newnode(f);   break;
		case SYM_NUL:   f_nul(f);       break;
		case SYM_POP:	f_pop(f);       break;
		case SYM_SEND:  f_send(f);      break;
		case SYM_SPLIT: f_split(f);     break;
		case SYM_COPY:  f_copy(f);      break;
		case SYM_WAIT:  f_wait(f);      break;
		case SYM_TRIGGER: appl_trigger(f); break;
		default:
			fprintf(stderr, "'%s' not implemented\n",
			       name_lookup(s));
			if (--(f->mult) == 0)
				f_free(f);
			break;
		}
		return ops+1;
	}
	f1 = n->mtab[s].first;
	f2 = n->otab[s].first;
	if (!f1 || !f2)
		return ops;
	DEBUG(1, printf("%05d  ", n->opcnt);
	      f_print(stdout, f1, 0); printf(", ");
	      f_print(stdout, f2, 0); printf(" --> "););
	      
	if ((f1->len + f2->len - 3) > 0) {
		buf = malloc(sizeof(*buf) + sizeof(symbol) * (f1->len + f2->len - 3));
		buf->refcnt = 0;
		buf->len = f1->len + f2->len - 3;
		memcpy(buf->data, f1->buf->data + f1->offs + 2, sizeof(symbol)*(f1->len - 2));
		memcpy(buf->data + f1->len - 2, f2->buf->data + f2->offs + 1, sizeof(symbol)*(f2->len - 1));
		f3 = f_new(buf, 0, buf->len, 1, f1->nd);
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, f_print(stdout, f3, 1););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", ");
			      f_print(stdout, f3, 1););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
		node_lookup(f3->nd, 0)->symbalance -= f3->len;
		rx_fraglet(f3);
	} else {
		switch (f1->buf->data[f1->offs]) {
		case SYM_MATCH:
			DEBUG(1, fprintf(stdout, "[]\n"););
			break;
		case SYM_MATCHP:
			DEBUG(1, f_print(stdout, f1, 0);
			      fprintf(stdout, ", []\n"););
			break;
		default:
			fprintf(stderr, "wrong MATCH fragment in list\n");
			break;
		}
	}

	n->mtab[s].first = f1->next;
	if (n->mtab[s].last == f1)
		n->mtab[s].last = 0;
	n->otab[s].first = f2->next;
	if (n->otab[s].last == f2)
		n->otab[s].last = 0;

// we could do better here: append directly, we know the front symbol s ...
	if (f1->buf->data[f1->offs] == SYM_MATCH) {
		node_lookup(f1->nd, 0)->symbalance += f1->len;
		f1->mult--;
	}
	if (f1->mult <= 0)
		f_free(f1);
	else
		rx_fraglet(f1);

	node_lookup(f2->nd, 0)->symbalance += f2->len;
	f2->mult--;
	if (f2->mult <= 0)
		f_free(f2);
	else
		rx_fraglet(f2);

	return ops+1;
}
#endif


void f_cleanup()
{
	struct node *n;

	DEBUG(9, printf("## cleanup\n"););
	for (n = nodelist; n; n = n->next) {
		int i;
		struct m_rec *mtab;
		struct o_rec *otab;

		for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
			f_freeLinkedList((struct t_queue *) mtab);
		for (i = 0, otab = n->otab; i < MAXSYMBOLS; i++, otab++)
			f_freeLinkedList((struct t_queue *) otab);
		f_freeLinkedList(&(n->tq));

		// (lidia): should free waiting list too!
		f_freeLinkedList(&(n->wait));

		n->opcnt = 0;
		n->r_active = n->t_active = 0;
		//		n->hintcnt = 0;
	}

}


void collect_hints(struct node *n)
{
	int i;
	struct m_rec *mtab;
	struct o_rec *otab;
	struct frag *f;

	//DEBUG(9, printf("## collecthints %s\n", name_lookup(n->name)););

	/*
	if (n->tq.first) {
		if (addhint(n, f_peek(n->tq.first)))
			return;
	}
	for (i = 0, mtab = n->mtab; i < MAXSYMBOLS; i++, mtab++)
		if (mtab->first && n->otab[i].first) {
			if (addhint(n, i))
				return;
		}
	*/
	n->r_active = n->t_active = 0;
	for (i = 0, mtab = n->mtab, otab = n->otab;
				     i < MAXSYMBOLS; i++, mtab++, otab++) {
		mtab->active_cnt = 0;
		for (f = mtab->first; f; f = f->next) {
			if (otab->first)
				mtab->active_cnt += f->mult;
		}

		otab->active_cnt = 0;
		for (f = otab->first; f; f = f->next) {
			if (mtab->first)
				otab->active_cnt += f->mult;
		}
		mtab->r_active_cnt = mtab->active_cnt;
		if (mtab->active_cnt > otab->active_cnt)
			mtab->r_active_cnt = otab->active_cnt;

		n->r_active += mtab->r_active_cnt;
//		n->o_active += mtab->r_active_cnt;
	}
	for (f = n->tq.first; f; f = f->next)
		n->t_active += f->mult;
	n->total_active = n->r_active + n->t_active;
}


void run(int lim)
{
	int i;
	struct node *n;
	symbol s;
	int clock = 0;

	// J. Zivanov: if lim <=0 then infinite loop
	// (for continuous packet processing)
	while (lim <= 0 || lim > clock) {
		// if (lim > 0) {
		//  DEBUG(9, printf("## run loop %d/%d\n", clock, lim););
		//}
		for (n = nodelist; n; n = n->next) {
		
			if (n == nodelist) listen_on_local_sockets();

			// no processing of remote nodes:
			if (n->location == Remote) continue;

		  	// pending: modify "wait" to take into accout
		  	// "real" time instead of virtual time

			if (n->wait.first && n->opcnt >= n->wait.first->expires) {
				struct frag *f = n->wait.first;
				n->wait.first = f->next;
				if (n->wait.last == f)
					n->wait.last = 0;
				f->next = 0;
				if (f->buf->data[f->offs] == SYM_WAIT)
				  f = f_shrink(f,1);
				else if (f->buf->data[f->offs] == SYM_DELAY)
				  f = f_shrink(f,2);
				else {
				  printf("** bad: unknown head in wait queue\n");
				  exit(-1);
				}
				DEBUG(1,
				      printf("%05d  ",
					     node_lookup(f->nd, 0)->opcnt);
				      printf("... --> ");
				      f_print(stdout, f, 1););

				if (f)
					rx_fraglet(f);
				n->possibleHint |= 1;
			}

			// J. Zivanov: (lidia:) why ??????????
			//n->possibleHint |= appl_clockEvent(n, clock);

			/*
			if (!n->hintcnt && n->possibleHint)
				collect_hints(n);
			if (n->hintcnt) {
				i = random() % n->hintcnt;
				s = n->hints[i];
				n->hintcnt--;
				if (i != n->hintcnt)
					n->hints[i] = n->hints[n->hintcnt];
				i = f_exec(n, s);
			} else
				n->possibleHint = 0;
			*/
			collect_hints(n);
			DEBUG(9, printf("## run loop %d/%d: node=%s tactive=%d twait=%lu\n", clock, lim, name_lookup(n->name), n->total_active, n->wait.first););


			if (n->total_active) {
				i = random() % n->total_active;
//				printf("i=%d from %d\n", i, n->total_active);
				if (i < n->r_active) {
					struct m_rec *mr;
					mr = n->mtab;
					while (i >= mr->r_active_cnt) {
//						printf("m\n");
						i -= mr->r_active_cnt;
						mr++;
					}
//					printf("picking mtab %d\n", i);
					i = random() % mr->active_cnt;
					f_execFromQueue(n, (struct t_queue *)mr, i);
				} else {
						i -= n->r_active;
						f_execFromQueue(n, &n->tq, i);
				}
//				}

			}
			
			n->opcnt++;
		}
		clock++;

	  // not good place: select ( timeout ) // udp socket ??? or > 1 socket

	}
}

int main(int argc, char *argv[])
{
	int lim = 100;
	int soffs = 0;

	init_sockets();

	for (argc--, argv++; argc > 0; argc--, argv++) {
		if (!strcmp(argv[0], "-d") && argc > 1) {
			debug = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-e") && argc > 1) {
			dump = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-lim") && argc > 1) {
			lim = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-soffs") && argc > 1) {
			soffs = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		if (!strcmp(argv[0], "-lose") && argc > 1) {
			packet_loss_probability = atoi(argv[1]);
			argc--;
			argv++;
			continue;
		}
		
		fprintf(stderr, "unknown option %s\n", argv[0]);
		fprintf(stderr, "usage: frag [options]\n");
		fprintf(stderr, "  -d lev     // debug\n");
		fprintf(stderr, "  -e N       // export: 1=in, 2=out, 3=both\n");
		fprintf(stderr, "  -lim N     // limit exec count\n");
		fprintf(stderr, "  -soffs N   // score index offset\n");
		fprintf(stderr, "  -lose N    // udp loss probability 1/N\n");
		exit(-1);
	}

	// (lidia) be careful to add symbols in the same order as frag.h
	// enum SYM_*
	name_add("nop");
	name_add("copy");
	name_add("delay");
	name_add("dup");
	name_add("empty");
	name_add("exch");
	name_add("lt");
	name_add("length");
	name_add("diff");
	name_add("sum");
	name_add("newname");
	name_add("newnode");
	name_add("mutate");
	name_add("nul");
	name_add("pop");
	name_add("send");
	name_add("split");
	name_add("wait");
	name_add("_");

	name_add("match");
	name_add("matchp");
	name_add("*");
	name_add("0"); // (lidia): placeholder for all numbers
	
	// SYM_EOL is end-of-list marker

	while (!f_import(stdin)) {

		appl_scoreReset();
		// Lidia: reset also seed for random(), for repeatability
		srandom(1);

		if (dump || debug) {
			time_t t;
			struct passwd *u;
			struct utsname un;

			time(&t);
			u = getpwuid(getuid());
			uname(&un);

			printf("# " VERSION);
			printf("# Executed by %s on %s at %s",
			       u->pw_name, un.nodename, ctime(&t));
		}

		if (dump == 1 || dump == 3) {
			printf("\n# - - - - INITIAL MULTISET: - - - - - - -\n");
			f_export(stdout);
		}
		if (debug)
			printf("\n# - - - - EXECUTION: - - - -- - - - - - -\n");

		run(lim);

		if (dump == 2 || dump == 3) {
			printf("\n# - - - - END MULTISET: - - - - - - - - -\n");
			f_export(stdout);
		}

		if (lim > 0) {
			if (dump || debug)
				printf("\n# - - - - - - - - - - -\n");
			//	printf("# score: %d\n", finalScore());
			printf("s %d %d\n", appl_finalScore(), ++soffs);
		}

		f_cleanup();
	}
	return 0;
}

#include "sockin.c"

// eof
